#include "Simulation.h"

//structs for the simulation
#include "SimulationStructs.h"

//Runge-Kutta-Fehlberg solver
#include "rkf45.hpp"

//the right hand side function for the rkf solver
#include "SimulationRHS.h"

//used to draw the lines
#include "DrawDebugHelpers.h" 

//the number of bodies in the system
 #define NR_OBJECTS 9
//#define NR_OBJECTS 2

// Sets default values
ASimulation::ASimulation()
{
	PrimaryActorTick.bCanEverTick = true;
}



//------------------------------------------------------------------------------------------------------------------------------------------
//INITIALIZING

void ASimulation::BeginPlay()
{
	Super::BeginPlay();
	if (useFixedTimeStep) timeScaleConstant = 1.0;
	InitializeBodies();
	tickCount = 0;
	simTime = 0.0;
	RecordSnapshot();
	InitializeSolver();
}


//initialize the gravitational bodies with their respective starting parameters
void ASimulation::InitializeBodies()
{
	//initialize the bodies for the selected problem
	RealSolarSystem();
	//SingleBodyTest();

	//move actors in scene to starting positions
	for (int i = 0; i < NR_OBJECTS; i++) actorRefs[i]->SetActorLocation(FVector(bodies[i].Pos.X, bodies[i].Pos.Y, bodies[i].Pos.Z)*100);
}


//initializes the bodies to simulate a single body at constant radius of 1
void ASimulation::SingleBodyTest()
{
	bodies.Add(FBody());
	bodies[0].Pos = FVectorDouble{ 0,0,0 };
	bodies[0].Vel = FVectorDouble{ 0,0,0 };	
	bodies[0].GM = 1.0;

	bodies.Add(FBody());
	bodies[1].Pos = FVectorDouble{ 1.0, 0.0, 0.0 };
	bodies[1].Vel = FVectorDouble{ 0.0, 1.0, 0.0 };
	bodies[1].GM = 1.0;
}


//initializes the bodies to simulate the real solar system
void ASimulation::RealSolarSystem()
{
	double gravConst = 6.67408e-11;
	double AU = 149597870700;

	//sol
	bodies.Add(FBody());
	bodies[0].Pos = FVectorDouble{ 0,0,0 };
	bodies[0].Vel = FVectorDouble{ 0,0,0 };	
	bodies[0].Mass = 1.98847e30;
	//mercury
	bodies.Add(FBody());
	bodies[1].Mass = 0.33011e24;
	//venus
	bodies.Add(FBody());
	bodies[2].Mass = 4.8675e24;
	//earth
	bodies.Add(FBody());
	bodies[3].Mass = 5.9724e24;
	//mars
	bodies.Add(FBody());
	bodies[4].Mass = 0.64171e24;
	//jupiter
	bodies.Add(FBody());
	bodies[5].Mass = 1898.19e24;
	//saturn
	bodies.Add(FBody());
	bodies[6].Mass = 568.34e24;
	//uranus
	bodies.Add(FBody());
	bodies[7].Mass = 86.813e24;
	//neptune
	bodies.Add(FBody());
	bodies[8].Mass = 102.413e24;

	//import positions and velocities from txt file
	FString importedString;
	FString positionFile = FPaths::ProjectConfigDir();
	positionFile.Append(TEXT("RealPositions/2000positions.txt"));
	FFileHelper::LoadFileToString(importedString, *positionFile);

	//parse the imported string into an array of strings
	TArray<FString> parsedString;
	importedString.ParseIntoArray(parsedString, TEXT(","), true); 

	//loop through the string aray, casting the strings to doubles, and setting parameters
	for (int i = 0; i < 8; i++) {
		//set positions directly to imported values (in AU)
		bodies[i + 1].Pos = FVectorDouble{
			FCString::Atod(*parsedString[i * 6 + 0]),
			FCString::Atod(*parsedString[i * 6 + 1]),
			FCString::Atod(*parsedString[i * 6 + 2])
		};
		//set velocities by converting imported values (in AU/day) to meters/second
		bodies[i + 1].Vel = FVectorDouble{
			(FCString::Atod(*parsedString[i * 6 + 3]) * AU) / 86400,
			(FCString::Atod(*parsedString[i * 6 + 4]) * AU) / 86400,
			(FCString::Atod(*parsedString[i * 6 + 5]) * AU) / 86400
		};
	}

	//scale the system
	//sqrt(G*M/AU) M is mass of sun
	double velocityScale = sqrt(gravConst * bodies[0].Mass / AU); 
	for (int i = 0; i < 9; i++) {
		bodies[i].Vel = bodies[i].Vel / velocityScale;
		// Divide masses by mass sun to create scaled gravitational parameters
		bodies[i].GM  = bodies[i].Mass / bodies[0].Mass;
	}
	//
	timeScale = sqrt(gravConst*bodies[0].Mass/(AU*AU*AU));
}


//set the parameters for the rkf solver
void ASimulation::InitializeSolver()
{
	//solver parameters
	flag = 1;
	t = 0.0;
	relerr = 100 * sqrt(r8_epsilon());
	abserr = 100 * sqrt(r8_epsilon());

	//the bodies GM values for the equation
	double GMs[NR_OBJECTS];

	//initial values for solver
	for (int i = 0; i < NR_OBJECTS; i++) {
		y[i * 6 + 0] = bodies[i].Pos.X;
		y[i * 6 + 1] = bodies[i].Pos.Y;
		y[i * 6 + 2] = bodies[i].Pos.Z;
		y[i * 6 + 3] = bodies[i].Vel.X;
		y[i * 6 + 4] = bodies[i].Vel.Y;
		y[i * 6 + 5] = bodies[i].Vel.Z;
		GMs[i] = bodies[i].GM;
	}

	//send parameters to the solver
	SetParams(GMs,timeScaleConstant);
	SimulationRHS(t, y, dydt);
}



//------------------------------------------------------------------------------------------------------------------------------------------
//TICK METHOD

// Called every frame
void ASimulation::Tick(float DeltaTime)
{
	Super::Tick(DeltaTime);

	//forces the fixed delta time if the option is selected
    if (useFixedTimeStep) DeltaTime = fixedTimeStep;

	//run the simulation using the selected solver
	switch (solver) {
		case SolverEnum::EULER:
			Euler(DeltaTime);
			break;
		case SolverEnum::EULER_CROMER:
			EulerCromer(DeltaTime);
			break;
		case SolverEnum::RUNGE_KUTTA:
			RungeKutta(DeltaTime);
			break;
		case SolverEnum::RUNGE_KUTTA_FEHLBERG:
			RungeKuttaFehlberg(DeltaTime);
			break;
	}

	//move the actors to their bodies location and draw line between old and new pos
	for (int i = 0; i < NR_OBJECTS; i++) {
		FVector newPos = FVector(bodies[i].Pos.X, bodies[i].Pos.Y, bodies[i].Pos.Z) * 100;
		DrawDebugLine(GetWorld(), actorRefs[i]->GetActorLocation(), newPos, FColor(255, 0, 0), true, -1, 0, 1);
		actorRefs[i]->SetActorLocation(newPos);
	}
}



//------------------------------------------------------------------------------------------------------------------------------------------
//SOLVER METHODS

void ASimulation::Euler(double DeltaTime)
{
	for (int i = 0; i < simLoopsPerTick; i++) {
		//update positions according to velocity, then update velocities
		for (int j = 1; j < NR_OBJECTS; j++) bodies[j].Pos += bodies[j].Vel * DeltaTime * timeScaleConstant;
		for (int j = 1; j < NR_OBJECTS; j++) bodies[j].Vel += CalculateAcceleration(bodies[j].Pos, j) * DeltaTime * timeScaleConstant;

		//record a snapshot of the simulation
		simTime += (DeltaTime * timeScaleConstant) / timeScale;
		RecordSnapshot();
	}
}


void ASimulation::EulerCromer(double DeltaTime)
{
	for (int i = 0; i < simLoopsPerTick; i++) {
		//update velocities, then update positions according to velocity
		for (int j = 1; j < NR_OBJECTS; j++) bodies[j].Vel += CalculateAcceleration(bodies[j].Pos, j) * DeltaTime * timeScaleConstant;
		for (int j = 1; j < NR_OBJECTS; j++) bodies[j].Pos += bodies[j].Vel * DeltaTime * timeScaleConstant;

		//record a snapshot of the simulation
		simTime += (DeltaTime * timeScaleConstant) / timeScale;
		RecordSnapshot();
	}
}


FVectorDouble ASimulation::CalculateAcceleration(FVectorDouble position, int bodyIndex)
{
	FVectorDouble acceleration = {0,0,0};
	for (int j = 0; j < NR_OBJECTS; j++) {
		if (j != bodyIndex) {
			//calculate square distance between bodies
			FVectorDouble differenceInPositions = position - bodies[j].Pos;
			double distanceBetween = sqrt((differenceInPositions.X * differenceInPositions.X)
				+ (differenceInPositions.Y * differenceInPositions.Y)
				+ (differenceInPositions.Z * differenceInPositions.Z));
			double squareDistance = distanceBetween * distanceBetween;

			//calculate direction to other body
			FVectorDouble forceDirection = (differenceInPositions / distanceBetween) * -1;

			//calculate acceleration towards other body
			acceleration += (forceDirection * (bodies[j].GM / squareDistance));	
		}			
	}
	return acceleration;
}


void ASimulation::RungeKutta(double DeltaTime)
{
	for (int i = 0; i < simLoopsPerTick; i++) {
		TArray<FVectorDouble> deltaVel;
		TArray<FVectorDouble> deltaPos;
		for (int j = 0; j < NR_OBJECTS; j++) {
			deltaVel.Add(FVectorDouble());
			deltaPos.Add(FVectorDouble());

			//Vectors for the four sub-steps, dydt position and velocity
			FVectorDouble K1v,K2v,K3v,K4v,K1r,K2r,K3r,K4r;
			
			//K1r = v
			K1r = bodies[j].Vel;
			//K1v = a(r)
			K1v = CalculateAcceleration(bodies[j].Pos, j);
			//K2r = v + K1v * h/2
			K2r = bodies[j].Vel + (K1v * (DeltaTime * timeScaleConstant) / 2);
			//K2v = a(r + (K1r * h/2))      
			K2v = CalculateAcceleration((bodies[j].Pos + (K1r * (DeltaTime / 2 * timeScaleConstant))) , j);
			//K2r = v + K2v * h/2
			K3r = bodies[j].Vel + (K2v * (DeltaTime * timeScaleConstant) / 2);
			//K3v = a(r + (K2r * h/2))
			K3v = CalculateAcceleration((bodies[j].Pos + (K2r * (DeltaTime / 2 * timeScaleConstant))), j);
			//K4r = v + K3v * h
			K4r = bodies[j].Vel + (K3v * (DeltaTime * timeScaleConstant));
			//K4v = a(r + (K3r * h))    
			K4v = CalculateAcceleration((bodies[j].Pos + (K3r * (DeltaTime * timeScaleConstant))), j);

			// NOTE - AS THIS IS THE DELTA IT IS "=" NOT "+="
			//v(i+1) = v(i) + h/6 * (K1v + 2*K2v + 2*K3v + K4v)
			deltaVel[j] = (K1v + (K2v * 2) + (K3v * 2) + K4v) * ((DeltaTime * timeScaleConstant) / 6);
			//r(i+1) = r(i) + h/6 * (K1r + 2*K2r + 2*K3r + K4r)
			deltaPos[j] = (K1r + (K2r * 2) + (K3r * 2) + K4r) * ((DeltaTime * timeScaleConstant) / 6);
		}

		//update velocities and positions according to the calculated vals
		for (int j = 1; j < NR_OBJECTS; j++) {
			bodies[j].Vel += deltaVel[j];
			bodies[j].Pos += deltaPos[j];
		}

		//record a snapshot of the simulation
		simTime += (DeltaTime * timeScaleConstant) / timeScale;
		RecordSnapshot();
	}
}


void ASimulation::RungeKuttaFehlberg(double DeltaTime)
{
	for (int i = 0; i < simLoopsPerTick; i++) {
		//run the solver
		tout = t + (DeltaTime);
		flag = r8_rkf45(SimulationRHS, NEQN, y, dydt, &t, tout, &relerr, abserr, flag);
		if (flag == 7) flag = 2;

		//update positions according to solver produced values
		for (int j = 0; j < NR_OBJECTS; j++) {
			bodies[j].Pos.X = y[j * 6 + 0];
			bodies[j].Pos.Y = y[j * 6 + 1];
			bodies[j].Pos.Z = y[j * 6 + 2];
			bodies[j].Vel.X = y[j * 6 + 3];
			bodies[j].Vel.Y = y[j * 6 + 4];
			bodies[j].Vel.Z = y[j * 6 + 5];
		}

		//record a snapshot of the simulation
		simTime += (DeltaTime * timeScaleConstant) / timeScale;
		RecordSnapshot();
	}
}



//------------------------------------------------------------------------------------------------------------------------------------------
//RECORD SNAPSHOT METHOD

void ASimulation::RecordSnapshot()
{
	snapshots.Add(FSnapshot());

	//record the positions of the planets
	for (int i = 1; i < NR_OBJECTS; i++) 
		snapshots[tickCount].positions.Add(FVectorDouble{bodies[i].Pos.X, bodies[i].Pos.Y, bodies[i].Pos.Z});

	//record the time
	snapshots[tickCount].time = simTime;

	tickCount++;
}



//------------------------------------------------------------------------------------------------------------------------------------------
//END OF PLAY DATA EXPORTING

//called when the simulation ends (esc key pressed)
void ASimulation::EndPlay(const EEndPlayReason::Type EndPlayReason)
{
	Super::EndPlay(EndPlayReason);

	//export for real solar system
	ValidateRealSystem();

	// export for single body
	//ExportSingleBodyErrors();
}


//export planetary positioning data
void ASimulation::ExportPlanetPositions(FString planetName, int planetIndex)
{
	
}


//calculate and export errors in the single body test
void ASimulation::ExportSingleBodyErrors()
{
	TArray<double> errors;
	for (int i = 0; i < tickCount; i++) {
		double distance = sqrt((snapshots[i].positions[0].X * snapshots[i].positions[0].X)
				+ (snapshots[i].positions[0].Y * snapshots[i].positions[0].Y)
				+ (snapshots[i].positions[0].Z * snapshots[i].positions[0].Z));
		errors.Add((distance - 1) * 100);
	}

	FString logFile = FPaths::ProjectConfigDir();
	logFile.Append(TEXT("GeneratedErrors/CircularOrbit/Errors.m"));
	FString stringToSave = "SingleBodyErr = [\n";
	for (int i = 1; i < tickCount; i++) {
		stringToSave.Append(FString::Printf(TEXT("%e"),snapshots[i].time));
		stringToSave.Append(" ");
		stringToSave.Append(FString::Printf(TEXT("%e"),errors[i]));
		stringToSave.Append("\n");
	}
	stringToSave.Append("];\n");
	stringToSave.Append("plot(SingleBodyErr(:,1), SingleBodyErr(:,2));\n");
	stringToSave.Append("xlabel('Time');\n");
	stringToSave.Append("ylabel('Error %');\n");
	stringToSave.Append("xlim([0,50]);\n");
	FFileHelper::SaveStringToFile(stringToSave, *logFile);
}


//export errors for the validation against real positions
void ASimulation::ValidateRealSystem()
{
	//variables for tracking the years
	int commonYearCount = 0;
	int leapYearCount = 0;
	int leapYearTracker = 3;
	int secondsInYear;
	bool isLeapYear;

	//error collections
	TArray<double> averageErrors;
	TArray<double> mercuryErrors;
	TArray<double>   venusErrors;
	TArray<double>   earthErrors;
	TArray<double>    marsErrors;
	TArray<double> jupiterErrors;
	TArray<double>  saturnErrors;
	TArray<double>  uranusErrors;
	TArray<double> neptuneErrors;

	//loop through the snapshots
	for (int i = 0; i < tickCount; i++) 
	{
		// if it's a gap year add the extra day
		if (leapYearTracker == 3) 
		{
			secondsInYear = 31622400;
			isLeapYear = true;
		} 
		else 
		{
			secondsInYear = 31536000;
			isLeapYear = false;
		}

		//if a year has passed
		if ((snapshots[i].time - ((leapYearCount * 31622400) + (commonYearCount * 31536000) + secondsInYear)) >=0 
		&& (leapYearCount + commonYearCount) < 20)
		{
			//iterate the year
			if (isLeapYear)
			{
				leapYearCount++;
				leapYearTracker = 0;
			} 
			else
			{
				commonYearCount++;
				leapYearTracker++;
			} 
			int currentYear = 2000 + leapYearCount + commonYearCount;
			
			//import the real position data for the start of the new year
			TArray<FVectorDouble> realPositions;
			FString yearString = FString::Printf(TEXT("%i"),currentYear);
			FString importedString;
			FString positionFile = FPaths::ProjectConfigDir();
			positionFile.Append(TEXT("RealPositions/"));
			positionFile.Append(yearString);
			positionFile.Append("positions.txt");
			FFileHelper::LoadFileToString(importedString, *positionFile);
			TArray<FString> parsedString;
			importedString.ParseIntoArray(parsedString, TEXT(","), true); 
			for (int j = 0; j < 8; j++) {
				realPositions.Add(FVectorDouble{
					FCString::Atod(*parsedString[j * 6 + 0]),
					FCString::Atod(*parsedString[j * 6 + 1]),
					FCString::Atod(*parsedString[j * 6 + 2])
				});
			}

			//calculate errors for individual planets
			mercuryErrors.Add(CalculateValidationError(snapshots[i].positions[0], realPositions[0]));
			venusErrors  .Add(CalculateValidationError(snapshots[i].positions[1], realPositions[1]));
			earthErrors  .Add(CalculateValidationError(snapshots[i].positions[2], realPositions[2]));
			marsErrors   .Add(CalculateValidationError(snapshots[i].positions[3], realPositions[3]));
			jupiterErrors.Add(CalculateValidationError(snapshots[i].positions[4], realPositions[4]));
			saturnErrors .Add(CalculateValidationError(snapshots[i].positions[5], realPositions[5]));
			uranusErrors .Add(CalculateValidationError(snapshots[i].positions[6], realPositions[6]));
			neptuneErrors.Add(CalculateValidationError(snapshots[i].positions[7], realPositions[7]));

			//calculate average error for whole system (sum of all / 8)
			int errIndex = leapYearCount + commonYearCount - 1; 
			// averageErrors.Add((mercuryErrors[errIndex] + venusErrors[errIndex]   + earthErrors[errIndex]
			// 	+ marsErrors[errIndex] + jupiterErrors[errIndex] + saturnErrors[errIndex]
			// 	+ uranusErrors[errIndex] + neptuneErrors[errIndex]) / 8);
			averageErrors.Add((venusErrors[errIndex]   + earthErrors[errIndex]
				+ marsErrors[errIndex] + jupiterErrors[errIndex] + saturnErrors[errIndex]  //excludes mercury from the average
				+ uranusErrors[errIndex] + neptuneErrors[errIndex]) / 7);
		}
	}

	//export errors
	ExportValidationErrors(mercuryErrors, "Mercury");
	ExportValidationErrors(venusErrors,   "Venus");
	ExportValidationErrors(earthErrors,   "Earth");
	ExportValidationErrors(marsErrors,    "Mars");
	ExportValidationErrors(jupiterErrors, "Jupiter");
	ExportValidationErrors(saturnErrors,  "Saturn");
	ExportValidationErrors(uranusErrors,  "Uranus");
	ExportValidationErrors(neptuneErrors, "Neptune");
	ExportValidationErrors(averageErrors, "Average");
}


double ASimulation::CalculateValidationError(FVectorDouble simPos, FVectorDouble realPos)
{
	//get magnitude of error (difference in position real vs simulated)
	FVectorDouble differenceInPositions = realPos - simPos;
	double errorMagnitude = sqrt((differenceInPositions.X * differenceInPositions.X)
		+ (differenceInPositions.Y * differenceInPositions.Y)
		+ (differenceInPositions.Z * differenceInPositions.Z));

	//get magnitude of real distance from sun
	double realMagnitude = sqrt((realPos.X * realPos.X) + (realPos.Y * realPos.Y) + (realPos.Z * realPos.Z));

	//error is errormagnitude/realmagnitude
	return errorMagnitude/realMagnitude;
}


void ASimulation::ExportValidationErrors(TArray<double> errors, FString name)
{
	//get the file path
	FString logFile = FPaths::ProjectConfigDir();
	logFile.Append(TEXT("GeneratedErrors/RealSystem/"));
	FString fileName = name + "Errors.m";
	logFile.Append(fileName);

	FString stringToSave = name + "Errs = [\n";
	stringToSave.Append("\n");
	stringToSave.Append(FString::Printf(TEXT("%i"),0));
		stringToSave.Append(" ");
		stringToSave.Append(FString::Printf(TEXT("%e"),0.0));
		stringToSave.Append("\n");
	for (int i = 0; i < errors.Num(); i++) {
		stringToSave.Append(FString::Printf(TEXT("%i"),(i+1)));
		stringToSave.Append(" ");
		stringToSave.Append(FString::Printf(TEXT("%e"),errors[i]));
		stringToSave.Append("\n");
	}
	stringToSave.Append("];\n");
	FString plotLine = "plot(" + name + "Errs(:,1)," + name + "Errs(:,2));";
	stringToSave.Append(plotLine);
	stringToSave.Append("xlabel('Time(Years)');\n");
	FString errorLabel = "ylabel('" + name + " Error %');\n";
	stringToSave.Append(errorLabel);
	FFileHelper::SaveStringToFile(stringToSave, *logFile);
}